require 'structured/type'
require 'structured/parser_stack'
require 'structured/nullable'

require 'structured/boolean'
require 'structured/dictionary'
require 'structured/integer'
require 'structured/list'
require 'structured/string'

module Structured
  module Struct
    # Alias these types here so these constant are part of the lookup path
    # when being used in our struct definitions.
    Boolean = Structured::Boolean
    String = Structured::String
    Integer = Structured::Integer
    Dictionary = Structured::Dictionary
    List = Structured::List
    Nullable = Structured::Nullable

    def self.included(base)
      base.extend(ClassMethods)
    end

    def initialize
      @attributes = {}
    end

    def generate(_context)
      attrs = self.class.attributes.select do |name, _|
        has_attribute?(name)
      end

      attrs.keys.map { |name| [name, public_send(name)] }.to_h
    end

    def read_attribute(name)
      @attributes.fetch(name) { self.class.attributes[name].default_value }
    end

    def has_attribute?(name)
      !@attributes[name].nil?
    end

    def write_attribute(name, value)
      @attributes[name] = value
    end

    module ClassMethods
      include Structured::Type

      def attributes
        @attributes ||= {}
      end

      def required_attribute(name, type)
        define_accessor(name.to_s)
        attributes[name.to_s] = Attribute.new(self, type)
      end

      def optional_attribute(name, type, default: Attribute::UNDEFINED)
        define_accessor(name.to_s)
        default = default == Attribute::UNDEFINED ? type.default_value : default
        attributes[name.to_s] = Attribute.new(self, type, default_value: default)
      end

      def ignore_attribute(name)
        attributes[name.to_s] = :ignore
      end

      def parse(yaml_hash, stack:, context: nil)
        parsed_instance = new
        parse_attributes(yaml_hash || {}, stack: stack, context: context).each do |attribute, value|
          parsed_instance.send(:"#{attribute}=", value)
        end

        parsed_instance
      end

      def new_with_defaults
        new_instance = new
        attributes.each do |name, attribute|
          next if attribute == :ignore
          new_instance.send(:"#{name}=", attribute.default_value) if attribute.has_default?
        end
        new_instance
      end

      def format_stack_frame_element(name)
        ".#{name}"
      end

      protected

      def define_accessor(name)
        attribute_accessors_module.send(:define_method, name) do
          read_attribute(name)
        end

        attribute_accessors_module.send(:define_method, "#{name}?") do
          has_attribute?(name)
        end

        attribute_accessors_module.send(:define_method, "#{name}=") do |value|
          write_attribute(name, value)
        end
      end

      def attribute_accessors_module
        self::AttributeAccessors ||= begin
          mod = Module.new
          include(mod)
          mod
        end
      end

      def parse_attributes(yaml_hash, allow_unknowns: false, stack:, context:)
        unless yaml_hash.is_a?(Hash)
          raise Structured::Errors::InvalidValue.new(stack,
            "#{stack} expects a hash, but found #{yaml_hash.class.name}")
        end

        result = {}
        errors = []
        attributes.each do |name, attribute|
          begin
            next if attribute == :ignore

            if yaml_hash.key?(name)
              result[name.to_sym] = attribute.type.parse(yaml_hash[name],
                stack: stack.push(name, attribute.type), context: context)
            elsif !attribute.has_default?
              raise Structured::Errors::MissingKey.new(stack,
                "#{stack} (#{type_name}) requires the following key: #{name}")
            end
          rescue Structured::Errors::Base => exception
            errors << exception
          end
        end

        unknown_keys = yaml_hash.keys - attributes.keys.map(&:to_s)
        unless allow_unknowns || unknown_keys.empty?
          unknown_keys.each do |key|
            errors << Structured::Errors::UnknownKey.new(stack,
              "#{stack} (#{type_name}) does not support a #{key} field")
          end
        end

        raise Structured::Errors::MultipleErrors.union(stack, errors) unless errors.empty?
        result
      end
    end

    class Attribute
      UNDEFINED = Object.new

      attr_reader :parent, :type

      def initialize(parent, type, default_value: UNDEFINED)
        @parent = parent
        @type = type
        @default_value = default_value
      end

      def has_default?
        @default_value != UNDEFINED
      end

      def default_value
        @default_value unless @default_value == UNDEFINED
      end
    end
    private_constant :Attribute
  end
end
